"""
Description
===========

This extension is best used only when testing the ``Flask`` application, and
can be initialied only in a ``pytest.fixture`` for example.

It modifies the test client to make it a :class:`FlaskTestClient`. The latter
will have direct access to session variables via :class:`FlaskTestSession` and
to a request response validation class via :class:`FlaskTestResponse`


Classes
=======

.. autoclass:: FlaskTestingExt
  :members:

.. autoclass:: FlaskTestClient
  :members:

.. autoclass:: FlaskTestResponse
  :members:

.. autoclass:: FlaskTestSession
  :members:

"""
import typing as t
from collections.abc import MutableMapping

from flask import Response
from flask.testing import FlaskClient

if t.TYPE_CHECKING:
    from _typeshed.wsgi import WSGIApplication


class FlaskTestSession(MutableMapping):
    """
    Lorem ipsum
    """

    # https://flask.palletsprojects.com/en/2.2.x/api/#flask.sessions.SessionInterface
    def __init__(self, client):
        """
        Lorem ipsum
        """
        self._client = client

    def _call(self, fn: t.Callable):
        with self._client.session_transaction() as session:
            return fn(session)

    def set(self, key, value) -> None:
        """
        Lorem ipsum
        """
        with self._client.session_transaction() as session:
            session[key] = value

    def get(self, key, default=None):
        """
        Lorem ipsum
        """
        return self._call(lambda s: s.get(key, default))

    def items(self):
        """
        Lorem ipsum
        """
        return self._call(lambda s: dict(s.items()))

    def clear(self) -> None:
        """
        Lorem ipsum
        """
        return self._call(lambda s: s.clear())

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __getitem__(self, key):
        return self._call(lambda s: s[key])

    def __delitem__(self, key):
        with self._client.session_transaction() as session:
            del session[key]

    def __str__(self):
        return self._call(lambda s: s.__str__())

    def __repr__(self):
        return self._call(lambda s: s.__repr__())

    def __iter__(self):
        return iter(self.items())

    def __len__(self):
        # pylint: disable=unnecessary-lambda
        return self._call(lambda s: len(s))

    def __contains__(self, key):
        return self._call(lambda s: key in s)


class FlaskTestResponse(Response):
    """
    Lorem Ipsum
    """

    # https://flask.palletsprojects.com/en/2.2.x/api/#flask.Response
    def expect(self, status_code=200, *, content_type=None):
        """
        Lorem Ipsum
        """
        if status_code is not None and not isinstance(status_code, int):
            assert False, f"Status code '{status_code}' must be None, or an integer"

        # content_type:
        #   None  = guess, and check the guess against what we recieved,
        #   *     = accept anything, do not guess
        #   "xzz" = exact match
        #
        # Verify status code
        if status_code:
            message = f"Expected client to return HTTP code {status_code}, received {self.status_code} instead"
            assert status_code == self.status_code, message

        # If 'X-Content-Type-Options' is set to nosniff, then content-type should be set
        # in the response, and unit test should check it.
        # TODO: assert the self.content_type is set
        if self.headers.get("X-Content-Type-Options") == "nosniff":
            pass

        # Get the real data
        data = self.data
        if self.is_json:
            # This is a json payload, expected content-type should be set to json
            # FIXME: if data is b'' then it can actually be text/plain, etc...
            data = self.json if data else {}
            if not content_type:
                content_type = "application/json"

        if content_type == "*":
            pass
        elif content_type:
            # Alter the content type for some types
            if content_type == "application/json" and self.status_code >= 400:
                content_type = "application/problem+json"

            # Verify content_type
            content = self.content_type.split(";")[0]
            message = f"Expecting '{content_type}' as content-type but got '{content}'"
            assert content_type == content, message

        if content_type in ["application/json", "application/problem+json"]:
            assert self.is_json, "Expecting data to be json, but it was not"

        return self


class FlaskTestClient(FlaskClient):
    """
    A test client class that inherits from `Flask`'s test client and also
    from ``Werkzeug``'s test client.

    See Also
    --------

    `FlaskClient <https://flask.palletsprojects.com/en/latest/api/#flask.Test.FlaskClient>`_

    """

    # https://flask.palletsprojects.com/en/latest/api/#flask.Test.FlaskClient
    #
    # Inherits from Flask's FlaskClient:
    #   .venv/lib/python3.9/site-packages/flask/testing.py
    # And from Werkzeug's Client:
    #   .venv/lib/python3.9/site-packages/werkzeug/test.py
    #
    def __init__(self, *args, **kwargs):
        """
        Lorem Ipsum
        """
        super().__init__(*args, **kwargs)
        self._default_headers = {}

    # pylint: disable=arguments-differ
    def get(self, url: str, **kwargs) -> FlaskTestResponse:
        """
        Call :meth:`request` with ``method`` set to ``GET``.
        """
        return self.request(url, method="GET", **kwargs)

    # pylint: disable=arguments-differ
    def post(self, url: str, **kwargs) -> FlaskTestResponse:
        """
        Call :meth:`request` with ``method`` set to ``POST``.
        """
        return self.request(url, method="POST", **kwargs)

    # TODO: add all methods from
    # .venv/lib64/./python3.9/./site-packages/werkzeug/test.py
    def request(self, url: str, method="GET", session=None, headers=None, **kwargs) -> FlaskTestResponse:
        """
        The test client makes requests to the application without running a live server.
        Flask-Gordon's client extends Flask's client which itself extends Werkzeug's client.

        follow_redirects:
        query_string
        headers
        data
        json
        """
        # https://flask.palletsprojects.com/en/2.2.x/Test/#sending-requests-with-the-test-client
        # https://werkzeug.palletsprojects.com/en/2.2.x/test/
        #
        # Create or edit the session
        for k, v in (session or {}).items():
            self.session.set(k, v)

        # Set the headers
        # py39: headers = self._default_headers | (headers or {})
        headers = {**self._default_headers, **(headers or {})}

        if method == "GET":
            return super().get(url, headers=headers, **kwargs)
        if method == "POST":
            return super().post(url, headers=headers, **kwargs)
        assert False, f"call method '{method}' not yet implemented"

    @property
    def session(self) -> FlaskTestSession:
        """
        Returns an instance of :class:`FlaskTestSession`
        """
        return FlaskTestSession(self)

    @property
    def default_headers(self):
        """
        Lorem Ipsum
        """
        return self._default_headers

    @default_headers.setter
    def default_headers(self, value):
        """
        Lorem Ipsum
        """
        self._default_headers = value


class FlaskTestingExt:
    """
    Extension that alters the ``Flask`` default response class and test client
    class.
    """

    def __init__(self, app: t.Optional["WSGIApplication"] = None):
        """
        Initialize ``app`` with extension: alter the default ``Flask``
        :meth:`response_class` and the :meth:`test_client_class`.

        Parameters
        ----------

        app: WSGIApplication

            The WSGI application to bind extension to.
        """
        if app is not None:
            self.init_app(app)

    def init_app(self, app: "WSGIApplication") -> "WSGIApplication":
        """
        Initialize ``app`` with extension: alter the default ``Flask``
        :meth:`response_class` and the :meth:`test_client_class`.

        Parameters
        ----------

        app: WSGIApplication

            The WSGI application to bind extension to.
        """
        app.response_class = FlaskTestResponse  # type: ignore[attr-defined]
        app.test_client_class = FlaskTestClient  # type: ignore[attr-defined]
        return app



"""
Description
===========

.. docs/ext/argparse/flask.rst
.. include:: argparse/flask.rst

.. docs/ext/argparse/gunicorn.rst
.. include:: argparse/gunicorn.rst


Usage
=====

.. program:: flask-dev

.. code-block:: python

  #!/usr/bin/env python3

  from flask import Flask
  from flask_gordon.ext import ArgParseExt

  flask = Flask(__name__)
  flask = ArgParseExt(flask)


Classes
=======

.. autoclass:: ArgParseExt
   :members: __init__, init_app

"""
import argparse
import os
import sys
import typing as t

from torxtools import xdgtools
from torxtools.argtools import is_not_dir
from torxtools.cfgtools import which
from torxtools.pathtools import cachedir, expandpath
from yaml import safe_load

from ._utils import deepmerge
from .defaults import CFGFILE_SEARCH_PATHS, make_config
from .flask import FlaskHelper
from .gunicorn import GunicornHelper


def _read(cfgfile: str) -> t.Dict[str, t.Any]:
    """
    Convenience function in order to be mocked.

    Parameters
    ----------
    cfgfile: str

        a single path representing a yaml file.

    Returns
    -------
    dict:

        a dictionary
    """
    with open(cfgfile, encoding="UTF-8") as fd:
        data = safe_load(fd)
    return data or {}


class ArgParseExt:
    def __init__(
        self,
        app: "FlaskApp" = None,
        middleware: str = "gunicorn",
        argv: t.List[str] = None,
        cfgfilename=None,
        cfgfilepaths=None,
        default_cfg: t.Dict[str, t.Any] = None,
        description="Flasket server",
        key: str = "SETTINGS",
        appname: str = None,
    ):
        """
        Parameters
        ----------
        app: FlaskApp

            A Flask application.

        middleware: str

            Middleware to use: 'gunicorn' for production, 'flask' for debug.

        argv: list[str]

            Uses :code:`sys.argv[1:]` if None. Use :code:`[]` if you desire to deactive argument parsing.

        cfgfilename: str

            Configuration filename to use. Typically this is your packagename with a 'yml' extension.

        cfgfilepaths: str

            Paths to search for in order to find the configuration file. Every value of the list must contain the placeholder '{cfgfilename}' such as :code:`["/home/jdoe/{cfgfilename}"]`

        default_cfg: dict, default: :meth:`flasket.defaults.default_configuration`

            Dictionary containing the defaults for the command line arguments and configuration.
            Passed value will be merged with the default factory configuration, command line arguments
            and the configuration file that was read.

        description: str

            Text description to use in argparse.

        key: str

            Keyname where to save settings, defaults to app.config["SETTINGS"]

        appname: str

            Application name. Used to figure out cachepath. Otherwise, defaults to sys.argv[0]
        """
        if app is not None:
            self.init_app(
                app=app,
                argv=argv,
                middleware=middleware,
                cfgfilename=cfgfilename,
                cfgfilepaths=cfgfilepaths,
                default_cfg=default_cfg,
                description=description,
                key=key,
                appname=appname,
            )

    def init_app(
        self,
        app: "Flask",
        middleware: str = "gunicorn",
        argv: t.List[str] = None,
        cfgfilename=None,
        cfgfilepaths=None,
        default_cfg: t.Dict[str, t.Any] = None,
        description="Flask server",
        key: str = "SETTINGS",
        appname: str = None,
    ):
        # Sets environment variables for XDG paths
        xdgtools.setenv()

        # Verify middleware exists
        middleware = middleware.lower().strip()
        if middleware not in ["flask", "gunicorn"]:
            raise ValueError('middleware argument must be in ["flask", "gunicorn"]')
        if middleware == "flask":
            helper = FlaskHelper()
        if middleware == "gunicorn":
            helper = GunicornHelper()

        # Prepare search path for configuration file. Disable it if we're looking
        # for a impossible (None) file name
        if not cfgfilename:
            cfgfilename = "flask.yml"
        if cfgfilepaths is None:
            cfgfilepaths = CFGFILE_SEARCH_PATHS
        cfgfilepaths = [e.format(cfgfilename=cfgfilename) for e in cfgfilepaths]

        # Create a default configuration from what was passed
        # and what we set. Other values are filtered
        defaults = make_config(default_cfg)

        # Parse arguments if they exist
        if argv is None:
            argv = sys.argv[1:]
            if appname is None:
                appname = os.path.basename(sys.argv[0])

        arguments = self._parse_arguments(
            argv=argv,
            cfgfilepaths=cfgfilepaths,
            helper=helper,
            defaults=defaults,
            description=description,
        )

        # Copy arguments over to an dict with server
        arguments = {k: v for k, v in arguments.items() if v is not None}
        cfgfile = arguments.pop("cfgfile", None)
        arguments = {"server": arguments}

        # Search for the configuration file
        if cfgfile is None and cfgfilepaths:
            # Search for cfgfile
            cfgfilepaths = [e.format(cfgfilename=cfgfilename) for e in cfgfilepaths]
            cfgfile = which(cfgfile, expandpath(cfgfilepaths))

        filedata = _read(cfgfile or "/dev/null") or {}

        # Merge in the reverse order of priority
        cfg = defaults
        cfg = deepmerge(cfg, filedata)
        cfg = deepmerge(cfg, arguments)
        cfg = helper.force_cfg(cfg)

        app.config[key] = cfg

        # Set cachdir from configuration or default
        path = cfg.get("server", {}).get("cachedir", {})
        app.cachedir = cachedir(appname, path)

        return app

    @staticmethod
    def _parse_arguments(*, argv, cfgfilepaths, helper, defaults, description):
        if not argv:
            return {}

        # argument_default=None does not set the default to None for boolean options,
        # so we'll specifically set default=None for those values
        #
        # Default values aren't actually added/set here, but in the FlasketSettings,
        # We only care about values that were specified.
        parser = argparse.ArgumentParser(
            description=description,
            argument_default=None,
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        # Create helptext for UI option
        b_ui = {True: "enabled", False: "disabled"}[defaults["server"]["ui"]]
        if cfgfilepaths:
            # Keep on two lines, otherwise line continuation will make
            # an obsure argparse bug appear
            helpmsg_cfgfile = "Use CFGFILE as configuration file, "
            helpmsg_cfgfile += (
                f"otherwise first file found in search path is used. (default search path: {cfgfilepaths})"
            )
        else:
            helpmsg_cfgfile = "Use CFGFILE as configuration file."

        # fmt: off
        parser.add_argument(
            "-l", "--listen", metavar="HOST",
            help=f'The ip to listen on (default: {defaults["server"]["listen"]})',
        )
        parser.add_argument(
            "-p", "--port", metavar="PORT", type=int,
            help=f'The port to listen on (default: {defaults["server"]["port"]})',
        )
        parser.add_argument(
         "-c", "--cfgfile", metavar="CFGFILE",
           help=helpmsg_cfgfile,
           type=is_not_dir,
        )
        parser.add_argument(
            "--ui", action="store_true", default=None,
            help=f"Enable the OpenAPI UI. Disable with --no-ui. (default: {b_ui})",
        )
        parser.add_argument(
            "--no-ui", action="store_false", default=None, dest="ui",
            help=argparse.SUPPRESS,
        )
        # fmt: on
        helper.add_arguments(defaults["server"], parser)
        args = parser.parse_args(argv)
        return vars(args)



r"""
Description
-----------

Options
-------

.. program:: flasket

.. option:: -l HOST, --listen HOST

   The ip to listen on (default: localhost)

.. option:: -p PORT, --port PORT

   The port to listen on (default: 8080)

.. option:: -c CFGFILE, --cfgfile CFGFILE

   Use CFGFILE as configuration file, otherwise first file found in
   search path is used. (default search path: (...))

.. option:: --ui, --no-ui

   Enable the OpenAPI UI. Disable with :option:`--no-ui`. (default: enabled)

.. option::  -w WORKERS, --workers WORKERS

   Number of thread workers. (default: 0. If 0, cpu to use is (cpu_count * 2) with a maximum
   of 8; if negative, cpu to use is (cpu_count * 2) with no maximum.)

.. option::  --pidfile FILE

   A filename to use for the PID file. (default: none)

Usage
-----

.. program:: flasket

.. code-block:: python

  #!/usr/bin/env python3

  from flasket import make_flasket_cli

  if __name__ == "__main__":
      flasket = make_flasket_cli(__file__, "gunicorn", cfgname="example.yml")
      flasket.run()

"""
import argparse
import typing as t


class GunicornHelper:
    @staticmethod
    def force_cfg(cfg: t.Dict["str", t.Any]) -> None:
        cfg["server"]["production"] = True
        cfg["server"]["debug"] = False
        return cfg

    @staticmethod
    def add_arguments(defaults: t.Dict["str", t.Any], parser: argparse.ArgumentParser) -> None:
        """
        Adds Gunicorn middleware specific command line arguments :option:`--workers`, :option:`--pidfile`
        to the ArgumentParser.

        Parameters
        ----------
        defaults:

            a dictionnary containing default values

        parser: argparse.ArgumentParser

            a parser on which `add_argument` will be called
        """
        workers = defaults["workers"]
        pidfile = defaults["pidfile"]
        if not pidfile:
            pidfile = "none"

        # fmt: off
        parser.add_argument(
            "-w", "--workers", metavar="WORKERS", type=int,
            help=f"Number of thread workers. (default: {workers}. If 0, cpu to use is (cpu_count * 2) with a maximum of 8; if negative, cpu to use is (cpu_count * 2) with 8 maximum.)",
        )
        parser.add_argument(
            "--pidfile", metavar="FILE", type=str,
            help=f"A filename to use for the PID file. (default: {pidfile})",
        )
        # fmt: on



"""
Description
===========

Usage
=====

.. code-block:: python

  #!/usr/bin/env python3

  from flask import Flask
  from flask_gordon.ext import CeleryExt

  flask = Flask(__name__)
  celery = CeleryExt(flask)


Classes
=======

.. autoclass:: Celery
   :members: init_app

"""
import typing as t

try:
    from celery import Celery, Task

    HAS_CELERY = True
except ImportError:
    HAS_CELERY = False


DEFAULT_CONFIGURATION = {
    "broker_url": "redis://localhost:6379/0",
    "result_backend": "redis://localhost:6379/0",
    "broker_connection_retry_on_startup": False,
}


class CeleryExt:
    def __init__(self):
        if not HAS_CELERY:
            raise NotImplementedError("CeleryExt requires celery[redis] package")

    def init_app(
        self,
        app: "Flask",
        configuration=None,
    ):
        """
        Parameters
        ----------
        app: FlaskApp

            A Flask application.

        configuration: dict

            If a configuration dictionnary is passed, it will be used.
            Otherwise, a configuration will be searched for in app.config["CELERY"].

        """

        # pylint: disable=abstract-method
        class FlaskTask(Task):
            def __call__(self, *args: object, **kwargs: t.Dict[str, t.Any]) -> object:
                with app.app_context():
                    return self.run(*args, **kwargs)

        celery_app = Celery(app.name, task_cls=FlaskTask)

        # Set the first configuration that we find
        if not configuration:
            for name in ["celery", "CELERY"]:
                if name in app.config:
                    configuration = app.config[name]
                    break

        # Defaults
        # cf: https://docs.celeryq.dev/en/stable/userguide/configuration.html#example-configuration-file
        configuration = configuration or {}
        configuration = {**DEFAULT_CONFIGURATION, **configuration}

        app.config["CELERY"] = configuration
        celery_app.config_from_object(configuration)

        celery_app.set_default()
        app.extensions["celery"] = celery_app

        return celery_app
